@@ -46,6 +46,7 @@ extern "C" {
 #if (PERL_VERSION >= 10)
 #   define FAST_IV 1
 #endif
+
 #define DEFAULT_MAX_RECUR_DEPTH 10000
 
 #if !defined(HAVE_CSNAPPY)
@@ -218,7 +219,11 @@ srl_build_decoder_struct(pTHX_ HV *opt, sv_with_hash *options)
 
     dec->ref_seenhash = PTABLE_new();
     dec->max_recursion_depth = DEFAULT_MAX_RECUR_DEPTH;
-    dec->max_num_hash_entries = 0; /* 0 == any number */
+     /* 0 == any number for array, hash & string */
+    dec->max_num_hash_entries = 0;
+    dec->max_num_array_entries = 0;
+    dec->max_string_length = 0;
+    dec->max_uncompressed_size = 0;
 
     SRL_RDR_CLEAR(&dec->buf);
     dec->pbuf = &dec->buf;
@@ -257,6 +262,18 @@ srl_build_decoder_struct(pTHX_ HV *opt, sv_with_hash *options)
         if ( val && SvTRUE(val) )
             dec->max_num_hash_entries = SvUV(val);
 
+        my_hv_fetchs(he,val,opt, SRL_DEC_OPT_IDX_MAX_NUM_ARRAY_ENTRIES);
+        if ( val && SvTRUE(val) )
+            dec->max_num_array_entries = SvUV(val);
+
+        my_hv_fetchs(he,val,opt, SRL_DEC_OPT_IDX_MAX_STRING_LENGTH);
+        if ( val && SvTRUE(val) )
+            dec->max_string_length = SvUV(val);
+
+        my_hv_fetchs(he,val,opt, SRL_DEC_OPT_IDX_MAX_UNCOMPRESSED_SIZE);
+        if ( val && SvTRUE(val) )
+            dec->max_uncompressed_size = SvUV(val);
+
         my_hv_fetchs(he,val,opt, SRL_DEC_OPT_IDX_DESTRUCTIVE_INCREMENTAL);
         if ( val && SvTRUE(val) )
             SRL_DEC_SET_OPTION(dec,SRL_F_DECODER_DESTRUCTIVE_INCREMENTAL);
@@ -338,6 +355,9 @@ srl_build_decoder_struct_alike(pTHX_ srl_decoder_t *proto)
     dec->ref_seenhash = PTABLE_new();
     dec->max_recursion_depth = proto->max_recursion_depth;
     dec->max_num_hash_entries = proto->max_num_hash_entries;
+    dec->max_num_array_entries = proto->max_num_array_entries;
+    dec->max_string_length = proto->max_string_length;
+    dec->max_uncompressed_size = proto->max_uncompressed_size;
 
     if (proto->alias_cache) {
         dec->alias_cache = proto->alias_cache;
@@ -402,13 +422,13 @@ srl_decode_into_internal(pTHX_ srl_decoder_t *origdec, SV *src, SV *header_into,
     dec = srl_begin_decoding(aTHX_ origdec, src, start_offset);
     srl_read_header(aTHX_ dec, header_into);
     if (expect_false( SRL_DEC_HAVE_OPTION(dec, SRL_F_DECODER_DECOMPRESS_SNAPPY) )) {
-        dec->bytes_consumed = srl_decompress_body_snappy(aTHX_ dec->pbuf, dec->encoding_flags, NULL);
+        dec->bytes_consumed = srl_decompress_body_snappy(aTHX_ dec->pbuf, dec->encoding_flags, NULL, dec->max_uncompressed_size);
         origdec->bytes_consumed = dec->bytes_consumed;
     } else if (expect_false( SRL_DEC_HAVE_OPTION(dec, SRL_F_DECODER_DECOMPRESS_ZLIB) )) {
-        dec->bytes_consumed = srl_decompress_body_zlib(aTHX_ dec->pbuf, NULL);
+        dec->bytes_consumed = srl_decompress_body_zlib(aTHX_ dec->pbuf, NULL, dec->max_uncompressed_size);
         origdec->bytes_consumed = dec->bytes_consumed;
     } else if (expect_false( SRL_DEC_HAVE_OPTION(dec, SRL_F_DECODER_DECOMPRESS_ZSTD) )) {
-        dec->bytes_consumed = srl_decompress_body_zstd(aTHX_ dec->pbuf, NULL);
+        dec->bytes_consumed = srl_decompress_body_zstd(aTHX_ dec->pbuf, NULL, dec->max_uncompressed_size);
         origdec->bytes_consumed = dec->bytes_consumed;
     }
 
@@ -869,6 +889,11 @@ SRL_STATIC_INLINE void
 srl_read_string(pTHX_ srl_decoder_t *dec, int is_utf8, SV* into)
 {
     UV len= srl_read_varint_uv_length(aTHX_ dec->pbuf, " while reading string");
+    /* Limit the maximum size of the string that we accept to whatever was configured */
+    if (expect_false( dec->max_string_length != 0 && len > dec->max_string_length )) {
+        SRL_RDR_ERRORf2(dec->pbuf, "Got input string with %u characters, but the configured maximum is just %u",
+                (int)len, (int)dec->max_string_length);
+    }
     if (expect_false(is_utf8 && SRL_DEC_HAVE_OPTION(dec, SRL_F_DECODER_VALIDATE_UTF8))) {
         /* checks for invalid byte sequences. */
         if (expect_false( !is_utf8_string((U8*)dec->buf.pos, len) )) {
@@ -956,7 +981,11 @@ srl_read_array(pTHX_ srl_decoder_t *dec, SV *into, U8 tag) {
         (void)SvUPGRADE(into, SVt_PVAV);
     }
 
-    if (len) {
+    /* Limit the maximum number of elements that we accept to whatever was configured */
+    if (expect_false( dec->max_num_array_entries != 0 && len > dec->max_num_array_entries )) {
+        SRL_RDR_ERRORf2(dec->pbuf, "Got input array with %u entries, but the configured maximum is just %u",
+                (int)len, (int)dec->max_num_array_entries);
+    } else if (len) {
         SV **av_array;
         SV **av_end;
 
@@ -1577,7 +1606,7 @@ srl_read_reserved(pTHX_ srl_decoder_t *dec, U8 tag, SV* into)
 SRL_STATIC_INLINE void
 srl_read_regexp(pTHX_ srl_decoder_t *dec, SV* into)
 {
-    SV *sv_pat= FRESH_SV();
+    SV *sv_pat= sv_2mortal(FRESH_SV());
     srl_read_single_value(aTHX_ dec, sv_pat, NULL);
     SRL_RDR_ASSERT_SPACE(dec->pbuf, 1, " while reading regexp modifer tag");
     /* For now we will serialize the flags as ascii strings. Maybe we should use
@@ -1614,6 +1643,8 @@ srl_read_regexp(pTHX_ srl_decoder_t *dec, SV* into)
         }
 #ifdef MODERN_REGEXP
         {
+            SV *referent;
+            SV tmp;
             /* This is ugly. We have to swap out the insides of our SV
              * with the one we get back from CALLREGCOMP, as there is no
              * way to get it to fill our SV.
@@ -1624,8 +1655,9 @@ srl_read_regexp(pTHX_ srl_decoder_t *dec, SV* into)
              */
 
             /* compile the regex */
-            SV *referent= (SV*)CALLREGCOMP(sv_pat, flags);
-            SV tmp;
+            referent= (SV*)CALLREGCOMP(sv_pat, flags);
+            if (!referent)
+                SRL_RDR_ERROR(dec->pbuf, "bad regexp pattern");
 
             /* make sure the SV came from us (it should) and
              * is bodyless */
@@ -1643,14 +1675,12 @@ srl_read_regexp(pTHX_ srl_decoder_t *dec, SV* into)
 
             if (SWAP_DEBUG) { warn("after swap:"); sv_dump(into); sv_dump(referent); }
 
-            SvREFCNT_dec(sv_pat); /* I think we need this or we leak */
             /* and now throw away the head we got from the regexp engine. */
             SvREFCNT_dec(referent);
         }
 #elif defined( TRANSITION_REGEXP )
         {
             REGEXP *referent = CALLREGCOMP(aTHX_ sv_pat, flags);
-            SvREFCNT_dec(sv_pat);
             sv_magic( into, (SV*)referent, PERL_MAGIC_qr, 0, 0);
             SvFLAGS( into ) |= SVs_SMG;
         }
@@ -1666,7 +1696,6 @@ srl_read_regexp(pTHX_ srl_decoder_t *dec, SV* into)
             pm.op_pmflags= flags;
 
             re= CALLREGCOMP(aTHX_ pat, pat + pat_len, &pm);
-            SvREFCNT_dec(sv_pat);
             sv_magic( into, (SV*)re, PERL_MAGIC_qr, 0, 0);
             SvFLAGS( into ) |= SVs_SMG;
         }
